#include <stdio.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <netdb.h>
#include <signal.h>

#define UTMPDEF_ONLY

#include "logftmp.c"
#include "custlocality.h"

extern void usage();

void printBps(bps)
long bps;
{
     if (bps <= 9999)	printf(" %4ld Bps ",bps);
     else		printf(" %4ld kBps",bps/1024);
}


/* Preparation for MMAP() usage */
static caddr_t memmap = NULL;
static char *mbuf   = NULL;
static int   our_offset = -1;
#ifdef	USE_MMAP_xx
extern	caddr_t mmap( /*caddr_t addr, size_t len, int prot, int flags, int fd, off_t off */);
#endif


int main(argc,argv)
int argc;
char *argv[];
{
     int fd;
     struct utmp *ut, *ubuf;
     struct summary *summary;
     struct hostent *remotehost;
     struct sockaddr_in s_in;
     struct stat st;
     char buffer[256];
     long nation_count[12];
     long bandwidth[12];
     long upload[12];
     int download[12];
     int total_upload = 0;
     int total_download = 0;
     int ents,i,rc,rrc,nation;
     int sums_only = 0;
     int one_line_sum = 0;
     int no_reversals = 0;
     int no_lock = 0;
     int long_form = 0;
     int summary_dump = 0;
     int c, pidrc;
     int buflen;
     int use_averages = 0;
     int querylock = 0;
     time_t now = time(NULL);

     extern char *optarg;
     extern int optind, opterr;
     extern char *malloc();

     buflen = MAX_FTMP * sizeof(struct utmp) + sizeof(struct summary) + 50;
     summary = (struct summary *)malloc(buflen);
     if (!summary) {
       perror("out of memory on malloc()");
       exit(2);
     }
     ubuf = (struct utmp *)((char *)summary + SUMMARY_SIZE);

     for (nation = 0; nation < 12; ++nation) {
       nation_count[nation] = 0;
       bandwidth[nation] = 0;
       download[nation] = 0;
       upload[nation] = 0;
     }

     while ((c = getopt(argc,argv, "1sn?lANQS")) != -1)
       switch (c)
	 {
       case '1':
	   one_line_sum = 1;
       case 's':
	   sums_only = 1;
	   break;
       case 'n':
	   no_reversals = 1;
	   break;
       case 'N':
	   no_lock = 1;
	   break;
       case 'l':
	   long_form = 1;
	   break;
       case 'A':
	   use_averages = 1;
	   break;
       case 'S':
	   summary_dump = 1;
	   break;
       case 'Q':
	   querylock = 1;
	   break;
       default:
	   usage();
	 }

     fd = open(FTMPFILE,O_RDWR);
     if (fd<0) {
	  printf("Can't open `%s'\n",FTMPFILE);
	  exit(1);
     }

#ifdef	USE_FCNTL
     if (querylock) {
       struct flock fl;
       fl.l_type   = F_WRLCK;
       fl.l_whence = SEEK_SET;
       fl.l_start  = 0;
       fl.l_len    = 0;
       rc = fcntl(fd,F_GETLK,(void*)&fl);
       if (fl.l_type == F_UNLCK)
	 printf("No lock\n");
       else
	 printf("Lock with PID %d\n",fl.l_pid);
     }
     {
       struct flock fl;
       fl.l_type   = F_WRLCK;
       fl.l_whence = SEEK_SET;
       fl.l_start  = 0;
       fl.l_len    = 0;
       rc = fcntl(fd,F_SETLK,(void*)&fl);
     }
#else
#ifdef USE_FLOCK
     if (!no_lock)
       rc = flock(fd,LOCK_EX);
     /* block until you get it */
#else
     if (!no_lock)
       rc = lockf(fd,F_LOCK,0);
     /* block until you get it */
#endif
#endif

#ifdef USE_MMAP
     {
       lseek(fd,buflen,0);    /* Write something at the END + 1.. */
       write(fd,&buflen,1);   /* makes sure there exists space for MAX_FTMP */
       lseek(fd,0,0);	      /* entries and thus avoiding SIGBUS.. */
       memmap = mmap(/*No fixed place*/ NULL, buflen,
		     PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);
       rrc = buflen;
       ut = (struct utmp *)memmap;
       ents = MAX_FTMP;
     }
#endif
#ifdef USE_SYSVIPC
     {
       int rc;
       key_t key = ftok(FTMPFILE,SHMPROJID);

       errno = 0;
       rc = shmget(key, buflen, 0644);
       if (rc == -1) {
	 fprintf(stderr,"** shmget(%d,%d,0644) rc=-1, errno=%d\n",key,rc,errno);
	 return;
       }
       memmap = (void*)shmat(rc, NULL, 0);
       rrc = buflen;
       ut = (struct utmp *)memmap;
       ents = MAX_FTMP;
     }
#endif
     if ((long)memmap == -1) {
       fstat(fd,&st);
       ut = (struct utmp *)malloc(st.st_size);
       if (ut == NULL) {
	 perror("malloc failed for ut");
	 exit (1);
       }
       rrc =  read(fd, (void*)ut, st.st_size);
       ents = rrc / sizeof(struct utmp);
     }

     if ((long)memmap != -1 && memmap != NULL) {
       memcpy((void*)summary, (void*)ut, buflen);
       ut = ubuf;
     }

#ifdef USE_MMAP
     if ((long)memmap != -1 && memmap != NULL)
       munmap(memmap,buflen);
#endif

     lseek(fd,0,SEEK_SET);
#ifdef	USE_FCNTL
     {
       struct flock fl;
       fl.l_type   = F_UNLCK;
       fl.l_whence = SEEK_SET;
       fl.l_start  = 0;
       fl.l_len    = 0;
       rc = fcntl(fd,F_SETLK,(void*)&fl);
     }
#else
#ifdef USE_FLOCK
     if (!no_lock)
       rc = flock(fd,LOCK_UN);
#else
     if (!no_lock)
       rc = lockf(fd,F_ULOCK,0);
#endif
#endif

     close(fd);

     for (i=0; i<=12; i++) {
       nation_count[i] = bandwidth[i] = 0;
     }

     if (!sums_only)
       printf ("Nation  D/L      Host                  Idle   Last rate  Avg rate   Total kB\n");

     --ut; /* one before.. */
     for(i=0; i<ents; ++i) {
       int printthis = 0;

       ++ut;

       if (ut->pid == 0) continue; /* empty slot */
       pidrc = 0;
       if (!sums_only)
	 pidrc = kill (ut->pid,0);
       if (pidrc != 0)
	 continue;


       switch (ut->locality) {
     case CUST_LOCAL:
	 if (!sums_only) printf ("Finland  ");
	 nation = 0;
	 break;
     case CUST_CAMPUS:
	 if (!sums_only) printf ("Pictures ");
	 nation=4;
	 break;
     case CUST_NORDUNET:
	 if (!sums_only) printf ("NORDUnet ");
	 nation = 1;
	 break;
     case CUST_FOREIGN:
	 if (!sums_only) printf ("Foreign  ");
	 nation=2;
	 break;
     case CUST_AREA:
	 if (!sums_only) printf ("ARPA-IP  ");
	 nation=3;
	 break;
     default:
	 if (!sums_only) printf ("Others   ");
	 nation=3;
	 break;
       }
       nation_count[nation]++;
       /* bytecnttime  is updated at EVERY time the bytes are counted..
	  therefore if it is older than SHORT_TIMEOUT seconds, don't count
	  them on the bandwidth.. */
       if (ut->bytecnttime + SHORT_TIMEOUT> now) {
	 if (long_form) {
	   bandwidth[nation] += ut->longBps;
	   if (!one_line_sum && ut->longBps < 0)
	     printthis = 1;
	 } else {
	   if (!one_line_sum && ut->lastBps < 0)
	     printthis = 1;
	   bandwidth[nation] += ut->lastBps;
	 }
       }

       if (ut->uploader) {
	 if (!sums_only) printf("U");
	 total_upload++;
	 upload[nation]++;
       } else if (ut->sending) {
	 if (!sums_only) printf ("D");
	 total_download++;
	 download[nation]++;
       } else {
	 if (!sums_only) printf (" ");
       }
       
       if (!sums_only && !no_reversals &&
	   (s_in.sin_addr.s_addr = inet_addr(ut->ut_host)) != -1) {
	 remotehost = gethostbyaddr(&s_in.sin_addr, sizeof(s_in.sin_addr), AF_INET);
	 if (remotehost != NULL) { 
	   if (strlen(remotehost->h_name) > 25) {
	     remotehost->h_name += (strlen(remotehost->h_name) - 25);
	   }
	   if (pidrc == 0)
	     printf (" %-25.25s", remotehost->h_name);
	   else {
	     sprintf(buffer, "%d:%s", ut->pid,remotehost->h_name);
	     printf (" %-25.25s", buffer);
	   }
	 } else {
	   if (pidrc == 0) {
	     sprintf(buffer, "[%s]", ut->ut_host);
	     printf (" %-25.25s", buffer);
	   } else {
	     sprintf(buffer, "[%d:%s]", ut->pid,ut->ut_host);
	     printf (" %-25.25s", buffer);
	   }
	 }
       } else {
	 if (printthis || !sums_only) {
	   if (pidrc == 0) {
	     sprintf(buffer, "[%s]", ut->ut_host);
	     printf ("+%-25.25s", buffer);
	   } else {
	     sprintf(buffer, "[%d:%s]", ut->pid,ut->ut_host);
	     printf ("+%-25.25s", buffer);
	   }
	 }
       }

       if (printthis || !sums_only) {
	 printf(" %3d sec", (now - ut->ut_stamp));
	 printBps(ut->lastBps);
	 printBps(ut->longBps);
	 printf(" %6d kB", ut->bytecnt);
	 if (long_form) {
	   printf("  %5d %s",ut->pid,ut->ut_name);
	 }
	 printf("\n");
       }
     }

     if (!sums_only) printf ("\n");

     if (!one_line_sum) {
       printf ("Summary:");
       printf ("\n   Finnish                %3d, bandwidth", nation_count[0]);
       printBps (bandwidth[0] );
       printf ("\n   NORDUnet               %3d, bandwidth", nation_count[1]);
       printBps (bandwidth[1] );
       printf ("\n   Foreign                %3d, bandwidth", nation_count[2]);
       printBps (bandwidth[2] );
       printf ("\n   Pictures               %3d, bandwidth", nation_count[4]);
       printBps (bandwidth[4] );
       printf ("\n   Others/ARPA-IP         %3d, bandwidth", nation_count[3]);
       printBps (bandwidth[3] );
       printf ("\n   Upload/Download        %d/%d\n", total_upload, total_download);

       nation_count[2] += nation_count[3];
       nation_count[2] += nation_count[4];
       bandwidth[2] += bandwidth[3]; bandwidth[3] = 0;
       bandwidth[2] += bandwidth[4]; bandwidth[4] = 0;

     } else {
       int bandwidthsum = 0;

       nation_count[2] += nation_count[3];
       nation_count[2] += nation_count[4];
       bandwidth[2] += bandwidth[3]; bandwidth[3] = 0;
       bandwidth[2] += bandwidth[4]; bandwidth[4] = 0;

       for (i=0;i<12;++i)
	 bandwidthsum += bandwidth[i];
       if (!use_averages)
	 printf ("Fi/No/Fo Logins: %d %d %d Senders: %d %d %d Rates: %d %d %d Bps\n",
		 nation_count[0], nation_count[1], nation_count[2],
		 download[0],     download[1],     download[2],
		 bandwidth[0],    bandwidth[1],    bandwidth[2]);
       else
	 printf ("Fi/No/Fo Logins: %d %d %d Senders: %d %d %d Rates: %d %d %d Bps\n",
		 nation_count[0], nation_count[1], nation_count[2],
		 summary->avgsenders[CUST_LOCAL]/*DIV?*/,
		 summary->avgsenders[CUST_NORDUNET],
		 summary->avgsenders[CUST_FOREIGN],
		 summary->AvgBpses[CUST_LOCAL],
		 summary->AvgBpses[CUST_NORDUNET],
		 summary->AvgBpses[CUST_FOREIGN]);
     }

     if (summary_dump) {
       int i;
       if (summary->filler != FILLER_CODE && summary->server_pid != 0) {
	 printf("**summary processor (pid=%d) version is not latest!\n",
		summary->server_pid);
	 kill(summary->server_pid, SIGKILL);
       }
       printf("summary_server(): pid=%d, update age=%ds, client count=%d\n",
	      summary->server_pid, now - summary->last_update,
	      summary->total_clients);
       for (i=CUST_low; i<=CUST_high; ++i)
	 if ((use_averages && (summary->AvgBpses[i] || summary->avgsenders[i])) ||
	     (!use_averages && (summary->Bpses[i] || summary->senders[i]))) {
	   switch (i) {
	     case CUST_LOCAL:
	         printf(" LOCAL   ");
		 break;
	     case CUST_CAMPUS:
		 printf(" PICTURES");
		 break;
	     case CUST_AREA:
		 printf(" ARPA-IP ");
		 break;
	     case CUST_COUNTRY:
		 printf(" COUNTRY ");
		 break;
	     case CUST_NORDUNET:
		 printf(" NORDUNET");
		 break;
	     case CUST_FOREIGN:
		 printf(" FOREIGN ");
		 break;
	     case CUST_EUROPE:
		 printf(" EUROPE  ");
		 break;
	   }
	   if (use_averages)
	     printf(": %6d Bps, %3d senders\n",
		    summary->AvgBpses[i] /*DIV?*/, summary->avgsenders[i] /*DIV?*/);
	   else
	     printf(": %6d Bps, %3d senders\n",
		    summary->Bpses[i], summary->senders[i]);
	 }
     }
     exit(0);
}

void
usage()
{
  printf("ftmptool [-1] [-n] [-s] [-S] [-A] [-l]\n\
   prints any moment the status of the ftp server operations\n\
   -1 : Summary on one liner form\n\
   -s : Summary per `nation count'\n\
   -S : Another summary dump\n\
   -l : Long form listing\n\
   -A : Use averages kept by the summary server\n\
   -n : Don't reverse IP-numbers to host names (faster)\n"
	 );
  exit(1);
}
