/* Group of routines with common (static) data:
	char my_last_store[MAXPATHLEN]

   store() for  STOR, STOU, and APPE
   delete() for DELE
   makedir() for MKD
   removedir() for RMD

   plus some utilities:
	gunique() (STOU from store()),  log_store() (from store())

   Anonymous user may only delete what he/she stored/created with
   previous command.

 */


#include "ftpd.c"

static char my_last_store[MAXPATHLEN] = "";
static int  valid_my_last_store = 0;

static char *gunique();
extern char *strrchr();

extern void log_store();

void
store(name, mode, unique)
char *name, *mode;
int unique;
{
	FILE *fout, *din;
	struct stat64 st;
	int (*closefunc)();
	int statrc;
	int nullsink;
	int my_file;
	int successfull = 0;
	char *s;
	char *oname = name;
	char tmpname[MAXPATHLEN];
	char *tname = tmpname;
	char newbuf[MAXPATHLEN];
	char canonic[MAXPATHLEN+2];

	nullsink = (strcmp(name,"/dev/null")==0);
	canonicpath(canonic,workpath,name);
	chdir(workpath);
	my_file = valid_my_last_store && 
		  strcmp(canonic,my_last_store) == 0; /* true, if I stored it! */
	statrc = lstat64(name, &st);
	
	if (!my_file && !nullsink && (guest || non_anonymous) && statrc == 0) {
	  reply(425,"File already exists. Won't do STORE");
	  return;
	}
	if (guest || non_anonymous) umask(0755);  /* ---rw--w-  -- for scripts! */

	if (unique && statrc == 0 &&
	    (name = gunique(name,newbuf)) == NULL)
	  return;

	/* Create a STOR.nnnnn -file for temporary storage */
	if (guest && *mode != 'a') {
	  strcpy(tmpname,name);
	  s = strrchr(tmpname,'/');
	  if (!s) s = tmpname -1;
	  *++s = 0; /* Zap whatever after last '/' */
	  s = tmpname + strlen(tmpname);
	  sprintf(s,"STOR.%d",(int)getpid());
	  tname = tmpname;
	} else {
	  tname = name;		/* Non-guest does not use tmp-store-file */
	}


	if (restart_point)
	  mode = "r+w";
	

	seteuid(user_uid);
	
	fout = fopen64(tname, mode);

	seteuid(0);
	
	closefunc = fclose;
	if (fout == NULL) {
	  perror_reply(553, name);
	  return;
	}
	if (restart_point) {
	  if (type == TYPE_A) {
	    register int i, n, c;

	    n = restart_point;
	    i = 0;
	    while (i++ < n) {
	      if ((c=getc(fout)) == EOF) {
		perror_reply(550, name);
		(*closefunc)(fout);
		return;
	      }
	      if (c == '\n')
		i++;
	    }	
	    /*
	     * We must do this seek to "current" position
	     * because we are changing from reading to
	     * writing.
	     */
	    if (fseek(fout, 0L, SEEK_CUR) < 0) {
	      perror_reply(550, name);
	      (*closefunc)(fout);
	      return;
	    }
	  } else if (lseek(fileno(fout), restart_point, SEEK_SET) < 0) {
	    perror_reply(550, name);
	    (*closefunc)(fout);
	    return;
	  }
	}

	din = dataconn(name, (off_t)-1, "r", nullsink);
	if (din == NULL) {
	  (*closefunc)(fout);
	  if (tname == tmpname)
	    unlink(tname); /* Busted, deletable: delete the temporary.. */
	  return;
	}
	successfull = 0;
	if (receive_data(din, fout, nullsink) == 0) {
	  if (tname == name || rename(tmpname,name) == 0) {
	    if (unique)
	      reply(226, "Transfer complete (unique file name:%s).", name);
	    else
	      reply(226, "Transfer complete.");
	    canonicpath(my_last_store,workpath,name);
	    chdir(workpath);
	    valid_my_last_store = 1;
	    successfull = 1;
	  } else if (unique) {
	    /* Ah well, original UNIQUE target name failed, regenerate!
	       Here the tmpname is deletable! */
	    if ((name = gunique(oname)) == NULL) {
	      unlink(tmpname);
	      return;
	    }
	    if (rename(tmpname,name)==0) {
	      reply(226, "Transfer complete (unique file name:%s).", name);
	      canonicpath(my_last_store,workpath,name);
	      chdir(workpath);
	      valid_my_last_store = 1;
	      successfull = 1;
	    }
	  } else {
	    lreply(452,"Can't make file (%s) unique, but",name);
	    lreply(452,"stored it under name (%s)",tmpname);
	    lreply(452,"Don't store anything else into here if you wish it to");
	    lreply(452,"remain intact -- (%s) is temporary file name,",tmpname);
	    reply (452,"and as such, it will be reused upon your next STORE cmd.");
	    canonicpath(my_last_store,workpath,tmpname);
	    chdir(workpath);
	    valid_my_last_store = 1;
	    successfull = 1;
	  }
	} else {
	  if (tname == tmpname)
	    unlink(tmpname);	/* Well, it failed.. delete the STOR.nnnn */
	}
	(void) fclose(din);
	data = -1;
	pdata = -1;
	if (guest && successfull) {
	  prettypath(canonic,workpath,name);
	  log_store(canonic,mode);
	  sprintf(tmpname,"%s.uploaderinfo",my_last_store);
	  fout = fopen64(tmpname,mode);
	  if (fout) {
	    fprintf(fout,"%s %d host=%s passwd=`%s' identuser=%s\n", canonic, byte_count, remotehostname, ftmppasswd, auth_username);
	    fclose(fout);
	  } else {
	    syslog(FTP_LOG_PRIORITY,
		   "store() could not create `%s'! err: %m",tmpname);
	  }
	}
}


void
log_store(name,mode)
char *name;
char *mode;
{
	/* Input in canonic format */
	AccessLogging("S %d %s %lld %s\n", getpid(), remotehostname, byte_count, name);
}


void
delete(name)
	char *name;
{
	struct stat64 st;
	int my_file = 0;
	char canonic[MAXPATHLEN+2];

	if (guest || non_anonymous) {
	  canonicpath(canonic,workpath,name);
	  chdir(workpath);
	  my_file = valid_my_last_store && 
		    strcmp(canonic,my_last_store) == 0; /* true, if I stored it! */
	}

	if (!my_file && (guest || non_anonymous)) {
	  reply(520,"Anonymous may only delete what it stored with previous STOR/STOU command..");
	  return;
	}
	
	if (lstat64(name, &st) < 0) {
		perror_reply(550, name);
		return;
	}
	if (S_ISDIR(st.st_mode)) {
		if (rmdir(name) < 0) {
			perror_reply(550, name);
			return;
		}
		goto done;
	}
	if (unlink(name) < 0) {
		perror_reply(550, name);
		return;
	}
done:
	ack("DELE");
}

/*
 * Generate unique name for file with basename "local".
 * The file named "local" is already known to exist.
 * Generates failure reply on error.
 */
char *
gunique(local,newbuf)
	char *local, *newbuf;
{
	struct stat64 st;
	char *cp = strrchr(local, '/');
	int count = 0;

	if (cp)
		*cp = '\0';
	if (lstat64(cp ? local : ".", &st) < 0) {
		perror_reply(553, cp ? local : ".");
		return((char *) 0);
	}
	if (cp)
		*cp = '/';
	(void) strcpy(newbuf, local);
	cp = newbuf + strlen(newbuf);
	*cp++ = '.';
	for (count = 1; count < 100; count++) {
		(void) sprintf(cp, "%d", count);
		if (lstat64(newbuf, &st) < 0)
			return(newbuf);
	}
	reply(452, "Unique file name cannot be created.");
	return((char *) 0);
}

void
makedir(name)
	char *name;
{
	if (guest || non_anonymous)
	  umask(0000);

	seteuid(user_uid);

	if (mkdir(name, 0777) < 0)
		perror_reply(550, name);
	else {
		reply(257, "MKD command successful.");
		if (guest || non_anonymous) {
			char tmpname[MAXPATHLEN];
			FILE *fout;

			sprintf(tmpname,"%s.uploaderinfo",name);
			umask(0222);
			fout = fopen64(tmpname,"w");
			if (fout) {
			  fprintf(fout,"MKDIR %s host=%s passwd=`%s' identuser=%s\n", name, remotehostname, ftmppasswd, auth_username);
			  fclose(fout);
			}

			canonicpath(my_last_store,workpath,name);
			chdir(workpath);
			valid_my_last_store = 1;
			AccessLogging("M %d %s %s %s\n", getpid(), remotehostname, workpath, name);

		}
	}

	seteuid(0);
}

void
removedir(name)
	char *name;
{
	int my_file = 0;
	char canonic[MAXPATHLEN+2];

	if (guest || non_anonymous) {
	  canonicpath(canonic,workpath,name);
	  chdir(workpath);
	  my_file = valid_my_last_store && 
		    strcmp(canonic,my_last_store) == 0; /* true, if I stored it! */
	}

	if (!my_file && (guest || non_anonymous)) {
	  lreply(550,"Anonymous user may not delete directories, unless it");
	  reply (550,"was created by previous command..");
	  return;
	}

	if (rmdir(name) < 0)
		perror_reply(550, name);
	else
		ack("RMD");
}
