/*
 * Copyright (c) 1985, 1988 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Berkeley.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/**  Necessary files:
  /etc/passwd  (not when on ANONYMOUS)
  /etc/group   (not when on ANONYMOUS)
  /etc/shadow  (some systems)
  /etc/ftpusers
  /etc/ftpmessage (transient customer messages)
  /etc/ftpclosed  (transient file)
  /etc/ftpblocked (read contents, close connection -- depends on locality..)
  /etc/nologin    (from Shutdown...)
  /bin/ls
  /bin/compress		(retrieve() -tricks!)
  /bin/uncompress	(ln /bin/compress /bin/uncompress)
  /bin/gtar		(retrieve() -tricks!)
  /bin/pwd  (some systems)
  /bin/grep	(for FIND - finder() routine)
  /dev/log (for syslog())
  /dev/null (some obscure tricks with FAST /dev/null)
  /usr/adm/ftp.log
  /usr/adm/ftmp  ( logftmp() )
  ~ftp/.badloginroot/ directory containing very limited view of
  		      material available in system.
  ~ftp/files/ALL_FILES  text file used for "FIND" primitive..
**/

#ifndef	__STDC__
#define const /* Non-ansi compiler doesn't know about const.. */
#endif

#ifdef	USG
#define	SHADOWPWD	/* To enable shadow password on USG */
#endif

/*
 * File containing login names
 * NOT to be used on this machine.
 * Commonly used to disallow uucp.
 */
#define	FTPUSERS	"/etc/ftpusers"
/*
 * Permanent extra text message for customers.
 */
#define CUSTMSGFILE	"/l/etc/ftpmessage"
/*
 * Greeting message for IC account
 */
#define ICMOTD		"etc/motd"
/*
 * Message shown to foreign archive users
 */
#define FOREIGNMSGFILE	"/l/etc/ftpmessage.foreign"
/*
 * Locality, max anonymous, max-send-rate, etc.. configuration file:
 */
#define _PATH_FTPD_CONFIG "/l/etc/ftpd.conf"
/*
 * What to tell when FTPD is closed
 */
#define FTPDCLOSED	"/etc/ftpclosed"
#define FTPDBLOCKED	"/etc/ftpblocked"
#define NOLOGIN		"/etc/nologin"
/*
 * Collection place of log file (not via SYSLOG which is unreliable.)
 */
#define ACCESSLOGDIR	"/l/log/ftp/raw" /* ".../yymm/yymmdd" */


#ifndef lint
static char copyright[] =
"@(#) Copyright (c) 1985, 1988 Regents of the University of California.\n\
 All rights reserved.\n";
#endif /* not lint */

#ifndef lint
static char sccsid[] = "@(#)ftpd.c	5.27.1.1	(Berkeley) 3/2/89";
#endif /* not lint */


/*
 * FTP server.
 */
#include <varargs.h>

#include <sys/types.h>
#ifdef	_POSIX_SOURCE
#include <unistd.h>
#endif
#include <dirent.h>
#ifdef	USG
#define direct dirent	/* cultural differences */
#include <sys/bsdtypes.h>
#define NBBY 8		/* Dangerous!  Do it only if byte has 8 bits! */
#endif
#ifdef	ISC
#include <net/errno.h>
#endif
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <fcntl.h>
#ifndef	USG
#include <sys/wait.h>
#endif

#ifdef SecureWare	/* OSF/1 Enchanged Security */
#include <sys/security.h>
#include <prot.h>
#endif

#ifdef	BSD
# define strrchr rindex
# define strchr  index
#endif

#include <netinet/in.h>

#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdio.h>
#include <signal.h>
#include <pwd.h>
#include <setjmp.h>
#include <netdb.h>
#include <errno.h>
#ifdef	USE_STRINGS
#include <strings.h>
#else
#include <string.h>
#endif
#include <syslog.h>

#include "syslog-vals.h"
#include "custlocality.h"

extern	int errno;
extern	int h_errno;
extern	char *sys_errlist[];
extern	int sys_nerr;
extern	char *crypt();
extern	char version[];
extern	char *home;		/* pointer to home directory for glob */
extern	char cbuf[];

#include <time.h>

extern	int check_closure(char *);
extern void end_login(char *);

extern void fatal(char *);
extern void reply(int code, char *fmt, ...);
extern void lreply(int code, char *fmt, ...);
extern void lreplys(int code, FILE *, char *title, char *prefix);
extern void ack(char *s);
extern void nack(char *s);
extern void yyerror(char *s);
extern void cwd(char *path);
extern void canonicpath(char wpath[MAXPATHLEN+1], char workpath[MAXPATHLEN+1], char *path);
extern void prettypath(char wpath[MAXPATHLEN+1], char workpath[MAXPATHLEN+1], char *path);
extern void pwd(void);
extern char *renamefrom(char *name);
extern void renamecmd(char *from, char *to);

extern void dolog(struct sockaddr_in *);
extern void dologout(int status);
extern void myoob(void);

extern void passive(int vers);
extern void perror_reply(int code, char *string);
extern void send_file_list(char *whichfiles);

extern void setproctitle(char *fmt, ...);
extern void appendproctitletail(char *fmt, ...);
extern void setproctitletail(char *fmt, ...);
extern int  deny_badasciixfer(int, char *);
extern void newer(char *date, char *path, int showlots);
extern int  check_newer(char *, struct stat64 *, int);
extern void findpathcmd(char *, const char *);
extern void site_lsdbg(void);

extern int logftmp(char *, char *, char *, int, int, int *);
extern int getsetftmpBps(int, off64_t, time_t, int, int, int, int *, int *);
extern int endftmp(void);


extern int  microsleep(long);

extern void lostconn(void);
extern void floaterr(void);
extern FILE *dataconn(char *, off64_t, char *, int);
extern void send_data(FILE*, FILE*, off64_t, int(*closefunc)(FILE*), int, char *, long);
extern int  receive_data(FILE *, FILE *, int);

extern void AccessLogging(char *fmt, ...);
extern void SetupLogging(void);
extern void end_login(char *);
extern void user(char *);
extern void CheckCustomerMessages(void);
extern char *OrderStr(int n);
extern void DisplayCustomerMessages(int code);
extern void pass(char *passwdin);


extern	FILE *ftpd_popen(char *, char *);
extern  FILE *ftpd_vopenl(char *, char *, char *, ...);
extern	int   ftpd_pclose(FILE *);
extern	char *getline(char *, int, FILE *);


#ifdef	MAIN
#define	FTP_NAMES
#endif
#include <arpa/ftp.h>
#include <arpa/inet.h>
#include <arpa/telnet.h>

struct	sockaddr_in ctrl_addr;
struct	sockaddr_in data_source;
struct	sockaddr_in data_dest;
struct	sockaddr_in his_addr;
struct	sockaddr_in pasv_addr;

#ifndef NBBY
#ifdef __linux__
# define NBBY 8
#endif
#endif

#ifdef	MAIN
/* When some system has these, do not create them again.. */
#if	!defined(__hpux) && !defined(BSD) && !defined(USG) && !defined(__linux__)
 char * typenames[] = {
	"",
	"Ascii",
	"Ebcdic",
	"Image",
	"local"
};
 char * formnames[] = {
	"",
	"non-print",
	"telnet",
	"carriage control"
};
 char * strunames[] = {
	"",
	"stream file (no records)",
	"record file",
	"page file"
};
 char * modenames[] = {
	"",
	"stream",
	"block",
	"compressed"
};
#endif /* (No) need to define FTP_NAMES data.. */
#else /* Not MAIN -- we need these as externs.. */
extern char * typenames[];
extern char * formnames[];
extern char * strunames[];
extern char * modenames[];
#endif /* Not MAIN */

#ifndef	MAIN
# define EXTERN extern
#else
# define EXTERN
#endif

EXTERN int	data;
EXTERN jmp_buf	errcatch, urgcatch;
#if !defined(CMASK) || CMASK == 0
#undef CMASK
#define CMASK 027
#endif
#ifdef	MAIN
int	defumask = CMASK;		/* default umask value */
int	logged_in = 0;
struct	passwd *pw = NULL;
uid_t	user_uid = 0;		/* Who the user is */
int	timeout = 900;		/* timeout after 15 minutes of inactivity */
int	maxtimeout = 7200;/* don't allow idle time to be set beyond 2 hours */
int	usedefault = 1;		/* for data transfers */
int	pdata = -1;		/* for passive mode */
off64_t	byte_count = 0;
off64_t	cumul_bytes = 0;
off64_t	cumul_cmd_bytes = 0;
int	captivity = 0;
int	restart_point = 0;
char	*ClassReadme = NULL;
#else	/* Not MAIN */
extern int	defumask;
extern int	logged_in;
extern struct	passwd *pw;
extern uid_t	user_uid;
extern int	timeout;
extern int	maxtimeout;
extern int	usedefault;
extern int	pdata;
extern off64_t	byte_count;
extern off64_t	cumul_bytes;
extern off64_t	cumul_cmd_bytes;
extern int	captivity;
extern int	restart_point;
extern char	*ClassReadme;
#endif
EXTERN int	debug;
EXTERN int	logging;
EXTERN int	guest;
EXTERN int	type;
EXTERN int	form;
EXTERN int	stru;			/* avoid C keyword */
EXTERN int	mode;
EXTERN int	transflag;
EXTERN off64_t	file_size;
EXTERN char	tmpline[7];
EXTERN char	hostname[MAXHOSTNAMELEN];
EXTERN char	remotehost[MAXHOSTNAMELEN];
EXTERN char	remotehostname[MAXHOSTNAMELEN];
EXTERN int	remotenameOK;
EXTERN int	AllowAccess;		/* Allow access anyway.. */
EXTERN time_t	login_time;
EXTERN int	non_anonymous;

/*
 * Timeout intervals for retrying connections
 * to hosts that don't accept PORT cmds.  This
 * is a kludge, but given the problems with TCP...
 */
#define	SWAITMAX	90	/* wait at most 90 seconds */
#define	SWAITINT	5	/* interval between retries */

#ifdef	MAIN
int	swaitmax = SWAITMAX;
int	swaitint = SWAITINT;
#else
extern int	swaitmax, swaitint;
#endif

#ifdef SETPROCTITLE
#ifdef	MAIN
char	**Argv = NULL;		/* pointer to argument vector */
char	*LastArgv = NULL;	/* end of argv */
char	proctitle[BUFSIZ];	/* initial part of title */
#else
extern char proctitle[BUFSIZ];	/* initial part of title */
extern char **Argv,*LastArgv;
#endif
#endif /* SETPROCTITLE */
extern char VirtualChdir[];

EXTERN char ftmppasswd[160];
EXTERN int Counts[CUST_high+1];

#ifdef	MAIN
int ftpcnt=0;
CUST_LOCALITY custlocality=CUST_LOCAL;
#else
extern int ftpcnt;
extern CUST_LOCALITY custlocality;
#endif

#ifdef	MAIN
int usecanonicpath = 0;
int no_lreply = 0;
#else
extern int usecanonicpath;
extern int no_lreply;
#endif
EXTERN char workpath[MAXPATHLEN +1];

EXTERN char ttyline[20];
EXTERN char gttyline[20];
EXTERN char unameline[20];

EXTERN int login_attempts;	/* number of failed login attempts */
EXTERN int unlimited;		/* Unlimited anonymous account ? */
EXTERN int askpasswd;		/* had user command, ask for passwd */
EXTERN int no_lreplyo;
EXTERN int uploader;		/* Non-zero == can not do RETR commands */

/* int maxsendrate; */		/* 1 000 000  for max.  bytes/sec.. */

#ifdef MAIN
/* Spontaneous reply text.  To be sent along with next reply to user */
char	*autospout = NULL;
int	autospout_free = 0;
#else
extern char	*autospout;
extern int	autospout_free;
#endif

#include "identuser.h"
# ifndef	AUTH_TIMEOUT
#  define	AUTH_TIMEOUT	5
# endif	/* AUTH_TIMEOUT */
#ifdef	MAIN
unsigned short	auth_flag = 0;
#else /* !MAIN */
extern unsigned short auth_flag;
#endif

EXTERN char	auth_username[MAXHOSTNAMELEN+1];

#ifdef __hpux	/* Some systems have WEIRD ideas about how things should
		   be done..   HPUX is lame, you should see IBM AIX.. */
# define seteuid(x)  setresuid((uid_t)-1,(uid_t)x, (uid_t)-1)
#endif

/*
 * Signal portability things for ZMailer
 *
 * By Matti Aarnio <mea@utu.fi> 1995
 *
 */

/* TO BE CALLED AFTER   #include "hostenv.h"  !!! */

#ifndef SIGNAL_TYPE
#define SIGNAL_TYPE void	/* Propably not globally portable.. */
#endif

#if	defined(SA_NOCLDSTOP)||defined(SA_ONSTACK)||defined(SA_RESTART)
	/* ================ POSIX.1 STUFF ================ */
#define SIGNAL_HANDLE(X,Y) \
	{	struct sigaction act;		\
		act.sa_handler = Y;		\
		act.sa_flags   = SA_NOCLDWAIT;	\
		sigemptyset(&act.sa_mask);	\
		sigaction (X, &act, NULL);	\
	}
#define SIGNAL_HANDLESAVE(X,Y,OLD) \
	{	struct sigaction act, oact;	\
		act.sa_handler = Y;		\
		act.sa_flags   = SA_NOCLDWAIT;	\
		sigemptyset(&act.sa_mask);	\
		sigemptyset(&oact.sa_mask);	\
		sigaction (X, &act, &oact);	\
		OLD = oact.sa_handler;		\
	}
#define SIGNAL_HOLD(SIG) \
	{	sigset_t	sigmask;	\
		sigemptyset(&sigmask);		\
		sigaddset(&sigmask,SIG);	\
		sigprocmask(SIG_BLOCK,&sigmask,NULL);	\
	}
#define SIGNAL_RELEASE(SIG) \
	{	sigset_t	sigmask;	\
		sigemptyset(&sigmask);		\
		sigaddset(&sigmask,SIG);	\
		sigprocmask(SIG_UNBLOCK,&sigmask,NULL);	\
	}
#define SIGNAL_IGNORE(SIG) \
	SIGNAL_HANDLE(SIG,SIG_IGN)
	
	
#else	/* No SIGACTION (POSIX) stuff */
#ifdef	SV_INTERRUPT	/* BSDism */
	/* ================ BSD 4.3+ (?) STUFF ================ */
#define	SIGNAL_HANDLE(X,Y)	\
	{	struct sigvec sv;			\
		sigvec(X, (struct sigvec *)NULL, &sv);	\
		sv.sv_handler = Y;			\
		sv.sv_flags |= SV_INTERRUPT;		\
		sigvec(X, &sv, (struct sigvec *)NULL);	\
	}
#define	SIGNAL_HANDLESAVE(X,Y,OLD)	\
	{	struct sigvec sv;			\
		sigvec(X, (struct sigvec *)NULL, &sv);	\
		OLD = sv.sv_handler;			\
		sv.sv_handler = Y;			\
		sv.sv_flags |= SV_INTERRUPT;		\
		sigvec(X, &sv, (struct sigvec *)NULL);	\
	}
#define SIGNAL_HOLD(SIG) \
	sigblock(sigmask(SIG))
#define SIGNAL_RELEASE(SIG) \
	sigblock((~(sigmask(SIG))) & sigblock(0))
#define SIGNAL_IGNORE(SIG) \
	SIGNAL_HANDLE(SIG,SIG_IGN)

#else	/* No SIGACTION (POSIX), nor SIGVEC stuff */
	/* ================ Everything  else.. ================ */
	/*         ...  Well, I think it is  SVR3 ...		*/
#define	SIGNAL_HANDLE(X,Y)	\
		signal(X,Y)
#define	SIGNAL_HANDLESAVE(X,Y,OLD)	\
		OLD=signal(X,Y)
#define SIGNAL_HOLD(SIG) \
		sighold(SIG)
#define SIGNAL_RELEASE(SIG) \
		sigrelse(SIG)
#define SIGNAL_IGNORE(SIG) \
		sigignore(SIG)

#endif
#endif
