/*

  logftmp() -- tricks to find out how many anonymous FTP sessoins are going on.

 * Copyright (c) 1988 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Berkeley.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * static char sccsid[] = "@(#)logwtmp.c	5.2 (Berkeley) 9/20/88";
 */

#ifndef lint
static char sccsid[] = "@(#)logftmp.c	5.2 (Berkeley) 9/22/88";
#endif /* not lint */

#ifndef  __ARGS
# ifdef __STDC__
#  define __ARGS(x) x
# else
#  define __ARGS(x) ()
# endif
#endif

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <syslog.h>
#include "syslog-vals.h"
#include <errno.h>
#include <pwd.h>
#include "custlocality.h"
#ifdef USE_MMAP
#include <sys/mman.h>			/* Using MMAP() stuff */
#endif
#ifdef USE_SYSVIPC
#include <sys/ipc.h>
#include <sys/shm.h>
#endif

#ifndef __
# ifdef __STDC__
#  define __(x) x
# else
#  define __(x) ()
# endif
#endif
extern void microsleep __((long));


#define	FTMPFILE	"/usr/adm/ftmp"

#define TIMEOUT_SECS	(999)		/* A bit longer than our configurable
					   timeouts.. (900 sec = 15 mins) */
#define SHORT_TIMEOUT	(60*4)		/* Four minutes */

struct utmp {	/* This is our own utmp... */
	char	ut_name[32];
	char	ut_host[32];
	char	ut_line[12];
	time_t	ut_time;
	time_t	ut_stamp;	/* Last timestamp.. */
	unsigned int	locality:8;
	unsigned int	uploader:8;
	unsigned int	sending:1;
	unsigned int	filler1:15;
	time_t  bytecnttime;
	long bytecnt;
	int lastBps;	/* short term average */
	int longBps;	/* long-term average..(whole file) */
	int throttlerate; /* Last value of throttle rate */
	pid_t pid;
	int filler5;
      };


#ifdef __GNUC__
# define VOLATILE __volatile__
#else
#ifdef __STDC__
# define VOLATILE volatile
#else
#endif
# define VOLATILE
#endif

#define SUMMARY_SIZE 1024 /* more than the struct below.. */
struct summary { /* This is summary data */
VOLATILE int	filler;
#define FILLER_CODE 0x800020		/* Change when structure, or
					   algorithm is altered... */
VOLATILE pid_t	server_pid;
VOLATILE time_t	last_update;		/* Last update of this data */
	int	total_clients;
	long	Bpses[CUST_high+1];	/* Bps     per cust. class */
	long	senders[CUST_high+1];	/* senders per cust. class */
	long	AvgBpses[CUST_high+1];	/* Longer term average */
	long	avgsenders[CUST_high+1];
	long	Bytes[CUST_high+1];	/* Byte summs since 00:00:00 each day */
	long	BytesAver[CUST_high+1]; /* Bytes/sec decaying averages */
};

#define SHMPROJID 'F'

#ifndef	UTMPDEF_ONLY

static int fd = -1;

#ifdef	USG
#define ftruncate(x,y) {}
#endif

extern int errno;
extern void *realloc();
extern void *malloc __ARGS((unsigned int));
extern time_t time __ARGS((time_t *));
extern void strncpy __ARGS((void *, void *, int));

extern int uploader;	/* ftpd.c main() */

static struct utmp *utp = NULL;

static int ut_filled = 0;
static struct utmp *uut = 0L;
static int uutsiz = 0;
static time_t next_purgerun = 0;

/* Preparation for MMAP() usage */
static caddr_t memmap = NULL;
static struct summary *summary = NULL;
static struct utmp *umap = NULL;
static int buflen = MAX_FTMP * sizeof(struct utmp) + SUMMARY_SIZE + 50 + 16*1024;
static int   our_offset = -1;
#if 0
extern	caddr_t mmap( /*caddr_t addr, size_t len, int prot, int flags, int fd, off_t off */);
#endif

static void check_summary_server __ARGS((void));
static struct utmp *get_ftmp_slot __ARGS((int fd));

pid_t mypid;

#ifdef USE_SYSVIPC
static int shem_id = -1;
#else
static int remap_shmem = 0;
#endif

int
logftmp(line, name, host, locality, refill, counts)
	char *line, *name, *host;
	int locality, refill;
	int *counts;
{
	int cnt = 0, i;
	struct stat stbuf;

	if (fd <0) {
	  fd = open(FTMPFILE, O_RDWR, 0);
	}
	if(fd < 0) return 0;

#ifdef USE_SYSVIPC
	if ( NULL == memmap ) {
	  int rc;
	  key_t key = ftok(FTMPFILE,SHMPROJID);

	  errno = 0;
	  rc = shmget(key, buflen, IPC_CREAT|IPC_EXCL|0644);
	  if (rc == -1 && errno == EEXIST) {
	    errno = 0;
	    rc = shmget(key, buflen, 0644);
	  }
	  if (rc == -1){
	    close(fd);
	    fd = -1;
	    return;
	  }
	  memmap = (void*)shmat(rc, NULL, 0);
	  if ((long)memmap == -1L) {
	    memmap = NULL;
	    close(fd);
	    fd = -1;
	    summary = NULL;
	    umap = NULL;
	    return;
	  }

	  /* Get a pointer to the summary structure. */
	  summary = (struct summary*)memmap;
	  /* If UMAP points to something, MMAP() was successfull! */
	  umap    = (struct utmp *)((char *)memmap + SUMMARY_SIZE);
#if 0	  
	  /* Aquire us one semaphore as well.. */
	  shem_id = semget(key, 1, IPC_CREAT|IPC_EXCL|0644);
	  if (shem_id == -1 && errno == EEXIST) {
	    errno = 0;
	    shem_id = semget(key,1,0644);
	  }
#endif
	}
#else /* Not SYSVIPC */
	if ( NULL == memmap ) {
    remap_ftmp:

	  fstat(fd,&stbuf);

	  errno = 0;
	  memmap = mmap(/*No fixed place*/NULL, stbuf.st_size,
			PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);
	  if ((long)memmap == -1) {
	    syslog(FTP_LOG_PRIORITY,"logftmp() mmap('%s') failed: %m",FTMPFILE);
	    summary = NULL;
	    memmap  = NULL;
	  } else {
	    /* Get a pointer to the summary structure. */
	    summary = (struct summary*)memmap;
	    /* If UMAP points to something, MMAP() was successfull! */
	    umap    = (struct utmp *)((char *)memmap + SUMMARY_SIZE);
	  }
	}
	remap_shmem = 0;
#endif /* SYSVIPC or not */

	if (utp == NULL && umap != NULL)
	  utp = get_ftmp_slot(fd);
#ifndef USE_SYSVIPC
	if (remap_shmem) {
	  munmap(memmap,stbuf.st_size);
	  memmap = NULL;
	  goto remap_ftmp;
	}
#endif

	if (utp == NULL) return 0;

	if( !ut_filled || refill ){
	  ut_filled = 1;
	  strncpy(utp->ut_name, name, sizeof(utp->ut_name)-1);
	  utp->ut_name[sizeof(utp->ut_name)-1] = 0;
	  strncpy(utp->ut_host, host, sizeof(utp->ut_host)-1);	
	  utp->ut_host[sizeof(utp->ut_host)-1] = 0;
	  utp->locality = locality;
	  utp->uploader = uploader;
	  time(&utp->ut_time);
	}
	if (line != NULL) {
	  strncpy(utp->ut_line, line, sizeof(utp->ut_line)-1);
	  utp->ut_line[sizeof(utp->ut_line)-1] = 0;
	}
	utp->sending  = 0;

	time(&utp->ut_stamp);

	if (counts) {
	  for (i=CUST_low; i<=CUST_high; ++i)
	    counts[i] = 0;

	  for (i = 0; i < MAX_FTMP; ++i) {
	    if (umap[i].pid > 1) {
	      counts[umap[i].locality] += 1;
	      ++cnt;
	    }
	  }
	}

	/* Check the existance of summary processor, if it can't be
	   found, or is too old -- over 30 secs since the last update,
	   fork() and become one! */
	check_summary_server();


	return cnt;
}

int
endftmp()
{
	extern int uploader;

	if (fd < 0 || utp == NULL) return 0;

	utp->sending  = 0;
	time(&utp->ut_time);
	time(&utp->ut_stamp);
	utp->pid = 0;
	close(fd);
	fd = -1;

	return 0;
}

static
#ifdef __GNUC__
__inline__
#endif
int utmpcmp(up1,up2)
     struct utmp *up1,*up2;
{
	return (up1->pid - up2->pid);
}

static int lock_fd(fd)
int fd;
{
	lseek(fd,0,SEEK_SET);
#ifdef	USE_FCNTL
	fl.l_type   = F_WRLCK;
	fl.l_whence = SEEK_SET;
	fl.l_start  = 0;
	fl.l_len    = 0;
	return fcntl(fd,F_SETLK,(void*)&fl);
#else
#ifdef USE_FLOCK
	return flock(fd,LOCK_EX);
#else
	return lockf(fd,F_LOCK,0);
	/* block until you get it */
#endif
#endif
}

static int unlock_fd(fd)
int fd;
{
	lseek(fd,0,SEEK_SET);
#ifdef	USE_FCNTL
	fl.l_type   = F_UNLCK;
	fl.l_whence = SEEK_SET;
	fl.l_start  = 0;
	fl.l_len    = 0;
	fcntl(fd,F_SETLK,(void*)&fl);
#else
#ifdef USE_FLOCK
	flock(fd,LOCK_UN);
#else
	lockf(fd,F_ULOCK,0);
#endif
#endif
	return 0;
}

static struct utmp *get_ftmp_slot(fd)
int fd;
{
#ifdef	USE_FCNTL
	struct flock fl;
#endif
	struct utmp *uup1,*uup2;
	int uutcnt = 0, cnt = 0, rc;
	struct stat buf;
	time_t now;
	int written = 0;
	int i, latest_free, do_lock;
	int seen_self = 0;
	pid_t mypid = getpid();

	if (!umap) return NULL;

	(void)time(&now);

	/* do_lock = (our_offset == -1) || (now > next_purgerun); */
	do_lock = 1;
	while(do_lock) {
	  if (lock_fd(fd) != 0) {
	    if(errno==EACCES
#ifdef	EWOULDBLOCK
	       || errno==EWOULDBLOCK
#endif
	       ) {
	      microsleep(1000000);
	    } else {
	      syslog(LOG_ERR,"locking of ftmp failed, fd=%d, err=%m",fd);
	      close(fd);	/* Failure ! */
	      fd = -1;
	      return 0;
	    }
	  } else break;
	}

	uut = NULL;
	our_offset = -1;
#ifndef USE_SYSVIPC
	fstat(fd,&buf); /* Failure is unthinkable.. */
	if (buf.st_size < buflen) {
	  /* Time to expand! */
	  int i;
	  for (i=buf.st_size; i < buflen; i += 16)
	    write(fd,"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0",16);
	  remap_shmem = 1;
	}
#endif
	for (i = 0, uut = umap; i < MAX_FTMP; ++i, ++uut) {
	  if (uut->pid == mypid ||
	      uut->pid == 0     ||
	      kill(uut->pid, 0) != 0) { /* Same pid, or no pid,
					   or no proc ? */
	    memset(uut,0,sizeof(struct utmp));
	    uut->pid = mypid;
	    our_offset = i;
	    break;
	  }
	}

	if (now > next_purgerun)
	  next_purgerun = now+4*60; /* Four minutes apart.. */

    unlock_exit:
	if (do_lock)
	  unlock_fd(fd);
	return uut;
}


int getsetftmpBps(sending,mybytecnt,mybytecnttime,longBps,myBps,throttlerate,Bpsmyclass,myclasssenders)
  int sending, longBps, myBps, throttlerate, *Bpsmyclass,*myclasssenders;
  off64_t mybytecnt;
  time_t mybytecnttime;
{
	struct utmp *uup2, *uup, ut2;
	int uutcnt = 0;
	struct stat buf;
	int i;
	int rc;
	time_t now;


	if (our_offset < 0) return 0; /* No place is known :-( */

	(void)time(&now);
	/* our_offset is known with & without mmap() */

	uutcnt = MAX_FTMP;
	uut = umap;
	uup = &uut[our_offset];

	/* process it! */

	if (Bpsmyclass)
	  *Bpsmyclass = 0;
	if (myclasssenders)
	  *myclasssenders = 0;
	uup2 = uut;
	rc = -1;
	--uup2;
	/* If pointer to OUR dataslot exists, update our data! */
	if (uup) {
	  /* Thats ours.. */
	  uup->bytecnt = mybytecnt/1024;
	  uup->bytecnttime = mybytecnttime;
	  uup->lastBps = myBps;
	  uup->sending = sending;
	  uup->longBps = longBps;
	  uup->throttlerate = throttlerate;
	  uup->ut_stamp = now;
	  rc = 0;
	}

	if (!Bpsmyclass || !myclasssenders) {
	  /* If no pointers to return values, just update
	     data in our slot; -- did it above -- if such
	     slot (uup->) exists. */
	} else {
	  /* Right; return-value pointers exist,
	     now if there exists summary structure.. */
	  if (summary && uup) {
	    /* All quick-access pointers exist */
	    if (Bpsmyclass)
	      *Bpsmyclass = summary->Bpses[uup->locality];
	    if (myclasssenders)
	      *myclasssenders = summary->senders[uup->locality];
	  } else {
	    /* No, no summary, just scan thru like before.. */
	    for(i=0, uup2 = uut; i<uutcnt; ++i, ++uup2) {
	      if (uup2 == uup) {
		/* Thats ours.. */
		uup2->bytecnt = mybytecnt/1024;
		uup2->bytecnttime = mybytecnttime;
		uup2->lastBps = myBps;
		uup2->sending = sending;
		uup2->longBps = longBps;
		uup2->throttlerate = throttlerate;
		uup2->ut_stamp = now;
		rc = 0;
	      }
	      if ((utp->locality == uup2->locality) &&
		  uup2->sending &&
		  ((uup2->bytecnttime + SHORT_TIMEOUT) > now)) {
		/* Same locality, and not ages old.. */
		if (Bpsmyclass)
		  *Bpsmyclass += uup2->lastBps;
		if (myclasssenders)
		  *myclasssenders += 1;
	      }
	    }
	  }
	}
	return 0;
}



/* ================================================================ */

static void
summary_server()
{
	/* Summary server:
	   This runs a loop every 5-10 seconds (or so), and calculates what
	   the Bpses and senders is to be for every user class.
	   */
	struct utmp *utp;
	int i, cnt;
	struct summary sum;
	int once = 1;
	time_t now;
	pid_t mypid = getpid();
	memset(&sum,0,sizeof(sum));

	summary->filler = FILLER_CODE; /* Change this when contents change.. */
	summary->server_pid = mypid;
	for (;;) {
	  if (summary->server_pid != mypid ||
	      summary->filler     > FILLER_CODE) {
	    _exit(1); /* Ok, kill myself if no data can be entered.. */
	  }
	  time(&now);
	  summary->last_update = now;
	  for (i=CUST_low;i<=CUST_high;++i) {
	    sum.Bpses[i] = 0;
	    sum.senders[i] = 0;
	  }
	  utp = uut;
	  cnt = 0;
	  for (i=0; i<MAX_FTMP; ++i, ++utp) {
	    if (utp->pid == 0) {
	      /* Empty slot! */
	      continue;
	    }
	    if (now > next_purgerun &&
		kill(utp->pid,0) != 0) {
	      syslog(FTP_LOG_PRIORITY,
		     "summary_server(): purging nonexistent: '%s', '%s', '%s', %d",
		     utp->ut_name,utp->ut_host,utp->ut_line,utp->pid);
	      utp->pid = 0;
	      continue;
	    }
	    if ((now > next_purgerun) &&
		(now > (utp->ut_stamp + TIMEOUT_SECS))) {
	      /* This bastard exists but has not had any xfer activity for a
		 while -- kill(pid,0) returned ok */
	      kill(utp->pid,9);
	      utp->pid = 0;
	      syslog(FTP_LOG_PRIORITY,
		     "summary_server() killed frozen FTP server process: '%s', '%s', '%s', Age: %ds, %d",
		     utp->ut_name,utp->ut_host,utp->ut_line,
		     utp->ut_time - now, utp->pid);
	      /* Do NOT zap him yet! 
		 memset(utp, 0, sizeof(struct utmp)); */
	      /* That was TIMEOUT_SECS old, updated at every
		 retrieve, and during data transport */
	      continue;		/* Outdated!    */
	    }
	    ++cnt;
	    if (utp->sending && (utp->bytecnttime + SHORT_TIMEOUT) > now) {
	      sum.senders[utp->locality] += 1;
	      sum.Bpses  [utp->locality] += utp->lastBps;
	    }
	  }
	  summary->total_clients = cnt;

#if 0 /* From Linux kernel -- scheduler */
/*
 * These are the constant used to fake the fixed-point load-average
 * counting. Some notes:
 *  - 11 bit fractions expand to 22 bits by the multiplies: this gives
 *    a load-average precision of 10 bits integer + 11 bits fractional
 *  - if you want to count load-averages more often, you need more
 *    precision, or rounding will get you. With 2-second counting freq,
 *    the EXP_n values would be 1981, 2034 and 2043 if still using only
 *    11 bit fractions.
 */
extern unsigned long avenrun[];		/* Load averages */

#define FSHIFT		11		/* nr of bits of precision */
#define FIXED_1		(1<<FSHIFT)	/* 1.0 as fixed-point */
#define LOAD_FREQ	(5*HZ)		/* 5 sec intervals */
#define EXP_1		1884		/* 1/exp(5sec/1min) as fixed-point */
#define EXP_5		2014		/* 1/exp(5sec/5min) */
#define EXP_15		2037		/* 1/exp(5sec/15min) */

#define CALC_LOAD(load,exp,n) \
	load *= exp; \
	load += n*(FIXED_1-exp); \
	load >>= FSHIFT;
#endif

	  for (i=CUST_low;i<=CUST_high;++i) {
	    long val;
	    val = (summary->AvgBpses[i] * 1981L +
		   sum.Bpses[i] * ((1 << 11) - 1981));
	    sum.AvgBpses[i] = val >> 11;
	    val = (summary->avgsenders[i] * 1981L +
		   sum.senders[i] * ((1 << 11) - 1981));
	    sum.avgsenders[i] = val >> 11;
	  }

	  memcpy(summary->Bpses,sum.Bpses,sizeof(sum.Bpses));
	  memcpy(summary->senders,sum.senders,sizeof(sum.senders));
	  memcpy(summary->AvgBpses,sum.AvgBpses,sizeof(sum.AvgBpses));
	  memcpy(summary->avgsenders,sum.avgsenders,sizeof(sum.avgsenders));

	  if (once)
	    syslog(FTP_LOG_PRIORITY,"summary_server() started, pid=%d",
		   getpid());
	  once = 0;
	  if (now > next_purgerun)
	    next_purgerun = now+3*60;	/* Three minutes apart.. */
	  microsleep(2000000);		/* Once in TWO seconds.. */
	} /* loop forever */
}


static void check_summary_server()
{
	int rc;
	int pid;
	time_t now;
	int do_lock = 1;

	time(&now);

	if (!summary) /* No pointer there, propably not in MMAP mode.. */
	  return;

	/* No summary server ?  Become one! */


	while(do_lock) {
	  if (lock_fd(fd) != 0) {
	    if(errno==EACCES
#ifdef	EWOULDBLOCK
	       || errno==EWOULDBLOCK
#endif
	       ) {
	      microsleep(1000000);
	    } else {
	      syslog(LOG_ERR,"locking of ftmp failed, fd=%d, err=%m",fd);
	      close(fd);	/* Failure ! */
	      fd = -1;
	      return;
	    }
	  } else
	    break;
	}

	if (summary->server_pid > 1)
	  rc = kill(summary->server_pid,0);
	else
	  rc = -1; /* Bad pid.. */

	if (summary->server_pid < 2 || rc != 0 ||
	    now > (summary->last_update+30)) {


	  syslog(LOG_ERR,"check_summary_server: rc=%d pid=%d",
		 rc,summary->server_pid);


	  if (summary->server_pid >= 2 && rc == 0 &&
	      now > (summary->last_update+30)) {
	    /* There is a process, it exists, it has been far too lazy!
	       KILL IT        */
	    kill(summary->server_pid, 9);
	    syslog(FTP_LOG_PRIORITY,
		   "check_summary_server() killed frozen server process: pid %d, Age %ds",
		   summary->server_pid,now - summary->last_update);
	    summary->server_pid = 0; /* No process anymore.. */
	    summary->last_update = 0;
	  }
	  pid = fork();
	  if (pid == 0) { /* Child! */
#ifdef	SETPROCTITLE
	    setproctitle("FTPD: summary_server()");
#endif
	    /* Close everything -- frees locks too.. */
	    for (rc=getdtablesize()-1; rc >= 0; --rc)
	      close(rc);
	    summary_server();
	  }
	  microsleep(200000 /* 0.2 seconds */);

	}
	if (do_lock)
	  unlock_fd(fd);
}

#endif /* UTMPDEF_ONLY */

