/* An interesting extention to FTP protocol by Phil Karn - KA9Q:

   XMD5 filepath   -- Calculate a MD5 hash of a file

   Return an error, or:  200 SP xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
   A 16 byte hash value in HEX.

   Hacked into Funic FTP-server by  <mea@nic.funet.fi>

*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <arpa/ftp.h>
#include "md5.h"

#define MD5BLOCK	64	/* Preferred MD5 block size */
extern int type;
extern int guest;

extern char *malloc();

/* Compute MD5 hash of local file */
static int
md5hasher(fp,hash,ascii)
FILE *fp;
char hash[16];
int ascii;
{
	MD5_CTX md;
	char *buf;
	int len;

	MD5Init(&md);

	if(ascii){
		/* Text file, hash canonical form (newlines = crlf) */
		char *cp;
		int c;

		cp = buf = malloc(MD5BLOCK);
		len = 0;
		while((c = fgetc(fp)) != EOF){
			if(c == '\n'){
				*cp++ = '\r';
				if(++len == MD5BLOCK){
					MD5Update(&md,buf,len);
					cp = buf;
					len = 0;
				}
			}
			*cp++ = c;
			if(++len == MD5BLOCK){
				MD5Update(&md,buf,len);
				cp = buf;
				len = 0;
			}
		}
		if(len != 0)
			MD5Update(&md,buf,len);
		free(buf);
	} else {
		buf = malloc(BUFSIZ);
		while((len = fread(buf,1,BUFSIZ,fp)) != 0){
			MD5Update(&md,buf,len);
		}
		free(buf);
	}
	MD5Final(&md);
	memcpy(hash,md.digest,16);
	return 0;
}


void
md5hash(filepath)
char *filepath;
{
	FILE *fil;
	struct stat sb;
	unsigned char hash[16];
	char strhash[33];
	int i;

	if (stat(filepath,&sb) < 0) goto nofile;
	if (!guest && !S_ISREG(sb.st_mode)) {
	  reply(550,"File %s is not a regular one",filepath);
	  return;
	}
	if (guest && !S_ISREG(sb.st_mode) && !(sb.st_mode & S_IROTH))





	  goto nofile; /* Deny existance of a file who isn't globally readable */
	fil = fopen(filepath,"r");
	if (!fil) goto nofile;
	
	md5hasher(fil,hash,type == TYPE_A);

	for (i=0; i<16; ++i)
		sprintf(strhash+i+i,"%02X",hash[i] & 0xFF);
	reply(200,strhash);
	
	fclose(fil);
	return;

nofile:	;
	reply(550,"No such file: %s",filepath);
	if (fil) fclose(fil);
	return;
}
