/*
 * Copyright (c) 1988 The Regents of the University of California.
 * All rights reserved.
 *
 * This code is derived from software written by Ken Arnold and
 * published in UNIX Review, Vol. 6, No. 8.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Berkeley.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * static char sccsid[] = "@(#)popen.c	5.7 (Berkeley) 9/1/88";
 */

#ifndef lint
static char sccsid[] = "@(#)popen.c	5.2 (Berkeley) 9/22/88";
#endif /* not lint */

#include <varargs.h>
#include <sys/types.h>
#ifdef	USG
#include <sys/bsdtypes.h>
#endif
#include <sys/signal.h>
#include <stdio.h>
#include <pwd.h>
#include <syslog.h>

#ifdef __hpux
# define seteuid(x)  setresuid((uid_t)-1,(uid_t)x, (uid_t)-1)
#endif


extern int builtin_ls(); /* Only do a GLOB + FORK, don't exec! */

/*
 * Special version of popen which avoids call to shell.  This insures noone
 * may create a pipe to a hidden program as a side effect of a list or dir
 * command.
 */
static uid_t *pids;
static int fds = 64; /* Umm.. */

extern uid_t user_uid;
extern int debug;
extern FILE *tmperr;

extern char **glob(), **copyblk(), *strtok(), *strrchr();

extern void *malloc();


FILE *
ftpd_popen(program, type)
	char *program, *type;
{
	register char *cp;
	FILE *iop;
	int argc, gargc, pdes[2], pid;
	char **pop, *argv[100], *gargv[1000], *vv[2];

	if (*type != 'r' && (*type != 'w' || type[1]))
		return(NULL);

	if (!pids) {
		if (!(pids =
		    (uid_t *)malloc((u_int)(fds * sizeof(uid_t)))))
			return(NULL);
		memset(pids, 0, fds * sizeof(uid_t));
	}
	if (pipe(pdes) < 0)
		return(NULL);

	fflush(stdout);	/* Flush out buffers! */
	fflush(stderr); /* We may switch over filenos 0,1,2 to a new files.. */
	/* Truncation at callers ! 
	if (tmperr) {
		fseek(tmperr,0,0);
		ftruncate(fileno(tmperr),0);
	}
	*/

	iop = NULL;

	/* NO vfork() HERE!  We do a LOT before we exec (if at all..) */

 	switch(pid = fork()) {
	case -1:			/* error */
		(void)close(pdes[0]);
		(void)close(pdes[1]);
		goto free;
		/* NOTREACHED */
	case 0:				/* child */

		if (tmperr) {
			dup2(fileno(tmperr),2);
		} else {
			close(2);
		}
		(void)seteuid(0);
		(void)setuid(user_uid);
		/* break up string into pieces */
		for (argc = 0, cp = program;; cp = NULL)
			if (!(argv[argc++] = strtok(cp, " \t\n")))
				break;

		/* glob each piece */
		gargv[0] = argv[0];
		for (gargc = argc = 1; argv[argc]; argc++) {
			if (!(pop = glob(argv[argc],10))) {
				/* globbing failed */
				vv[0] = argv[argc];
				vv[1] = NULL;
				pop = copyblk(vv);
			}
			argv[argc] = (char *)pop;
			/* save to free later */
			while (*pop && gargc < 1000)
				gargv[gargc++] = *pop++;
		}
		gargv[gargc] = NULL;

		if (*type == 'r') {
			if (pdes[1] != 1) {
				dup2(pdes[1], 1);
				(void)close(pdes[1]);
			}
			(void)close(pdes[0]);
		} else {
			if (pdes[0] != 0) {
				dup2(pdes[0], 0);
				(void)close(pdes[0]);
			}
			(void)close(pdes[1]);
		}
		cp = strrchr(gargv[0],'/');
		cp = cp ? cp+1 : gargv[0];
		if (strcmp(cp,"ls")==0) {
		  exit(builtin_ls(gargc,gargv));
		}
		execv(gargv[0], gargv);
		_exit(1);
	}
	/* parent; assume fdopen can't fail...  */
	if (*type == 'r') {
		iop = fdopen(pdes[0], type);
		(void)close(pdes[1]);
	} else {
		iop = fdopen(pdes[1], type);
		(void)close(pdes[0]);
	}
	pids[fileno(iop)] = pid;

free:	return(iop);
}

FILE *
ftpd_vpopenl(path, type, va_alist)
	char *path;
	char *type;
	va_dcl
/*  ( char *path, char *type, char *arg0, char *arg1, ... , NULL ) */
{
	va_list ap;
	FILE *iop;

	int argc, pdes[2], pid;
#define	MAXARGS 200
	char *args[MAXARGS+2];
	extern char **environ;

	va_start(ap);
	argc = 0;
	while((args[argc++] = va_arg(ap, char *)) != NULL && argc < MAXARGS) ;
	args[argc++] = NULL;
	args[argc++] = NULL;
	args[argc++] = NULL;
	va_end(ap);

	if (*type != 'r' && (*type != 'w' || type[1]))
		return(NULL);

	if (!pids) {
		if (!(pids =
		    (uid_t *)malloc((u_int)(fds * sizeof(uid_t)))))
			return(NULL);
		memset(pids, 0, fds * sizeof(uid_t));
	}
	if (pipe(pdes) < 0)
		return(NULL);

	iop = NULL;

 	switch(pid = fork()) {
	case -1:			/* error */
		(void)close(pdes[0]);
		(void)close(pdes[1]);
		return (iop);
		/* NOTREACHED */
	case 0:				/* child */
		if (tmperr) {
			fseek(tmperr,0,0);
			dup2(fileno(tmperr),2);
			ftruncate(fileno(tmperr),0);
		} else {
			close(2);
		}
		seteuid(0);
		setuid(user_uid);
		if (*type == 'r') {
			if (pdes[1] != 1) {
				dup2(pdes[1], 1);
				(void)close(pdes[1]);
			}
			(void)close(pdes[0]);
		} else {
			if (pdes[0] != 0) {
				dup2(pdes[0], 0);
				(void)close(pdes[0]);
			}
			(void)close(pdes[1]);
		}
		execv(path, args);
		_exit(1);
	}
	/* parent; assume fdopen can't fail...  */
	if (*type == 'r') {
		iop = fdopen(pdes[0], type);
		(void)close(pdes[1]);
	} else {
		iop = fdopen(pdes[1], type);
		(void)close(pdes[0]);
	}
	pids[fileno(iop)] = pid;

	return(iop);
}


int
ftpd_pclose(iop)
	FILE *iop;
{
	register int fdes;
	long omask;
	int pid;	
	extern int subproc_stat_loc; /* in ftpd.c */
	

	/*
	 * pclose returns -1 if stream is not associated with a
	 * `popened' command, or, if already `pclosed'.
	 */
	if (pids[fdes = fileno(iop)] == 0)
		return(-1);
	(void)fclose(iop);
#if	defined(USG) || defined(SVR4) || defined(SYSV) || defined(__linux__)
	signal(SIGINT,SIG_IGN);
	signal(SIGQUIT,SIG_IGN);
	signal(SIGHUP,SIG_IGN);
	while ((pid = wait(&subproc_stat_loc)) != pids[fdes] && pid != -1);
	signal(SIGINT,SIG_DFL);
	signal(SIGQUIT,SIG_DFL);
	signal(SIGHUP,SIG_DFL);
#else
	omask = sigblock(sigmask(SIGINT)|sigmask(SIGQUIT)|sigmask(SIGHUP));
	while ((pid = wait(&subproc_stat_loc)) != pids[fdes] && pid != -1);
	(void)sigsetmask(omask);
#endif
	pids[fdes] = 0;
	return(subproc_stat_loc);
}
